/****************************************************************************
 * The main entry point and related JWM functions.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

#ifdef USE_SHAPE
int shapeEvent;
#endif

static char *CONFIG_FILE = "/.jwmrc";

static void Initialize();
static void EventLoop();
static void Shutdown();
static void HandleExit();
static void DoExit(int code);

static char *configPath = NULL;
static char *displayString = NULL;

/****************************************************************************
 ****************************************************************************/
int main(int argc, char *argv[]) {
	char *temp;
	int x;

	StartDebug();

	temp = getenv("HOME");
	if(temp) {
		configPath = Allocate(strlen(temp) + strlen(CONFIG_FILE) + 1);
		strcpy(configPath, temp);
		strcat(configPath, CONFIG_FILE);
	} else {
		configPath = Allocate(strlen(CONFIG_FILE) + 1);
		strcpy(configPath, CONFIG_FILE);
	}

	for(x = 1; x < argc; x++) {
		if(!strcmp(argv[x], "-v")) {
			DisplayAbout();
			DoExit(0);
		} else if(!strcmp(argv[x], "-h")) {
			DisplayHelp();
			DoExit(0);
		} else if(!strcmp(argv[x], "-p")) {
			ParseConfig(configPath);
			DestroyRootMenu();
			ReleaseFonts();
			ReleaseColors();
			DoExit(0);
		} else if(!strcmp(argv[x], "-display") && x + 1 < argc) {
			displayString = argv[++x];
		} else {
			DisplayUsage();
			DoExit(1);
		}
	}

	do {
		shouldExit = 0;
		shouldRestart = 0;
		ParseConfig(configPath);
		Initialize();
		EventLoop();
		Shutdown();
	} while(shouldRestart);
	Release(configPath);
	StopDebug();

	return 0;
}

/****************************************************************************
 ****************************************************************************/
void DoExit(int code) {
	if(configPath) {
		Release(configPath);
	}
	StopDebug();
	exit(code);
}

/****************************************************************************
 ****************************************************************************/
void EventLoop() {
	XEvent event;

	while(!shouldExit) {

		WaitForEvent(&event);

		switch(event.type) {
		case ButtonPress:
			HandleButtonPress(&event.xbutton);
			break;
		case KeyPress:
			HandleKeyPress(&event.xkey);
			break;
		case EnterNotify:
			HandleEnterNotify(&event.xcrossing);
			break;
		case LeaveNotify:
			HandleLeaveNotify(&event.xcrossing);
			break;
		case MotionNotify:
			HandleMotionNotify(&event.xmotion);
			break;
		case KeyRelease:
		case ButtonRelease:
			break;
		default:
			Debug("Unknown event type: %d", event.type);
			break;
		}

	}
}

/****************************************************************************
 ****************************************************************************/
void Initialize() {
	XSetWindowAttributes attr;
	XModifierKeymap *modmap;
	KeyCode altCode;
	int temp;
	int x;

	initializing = 1;
	display = XOpenDisplay(displayString);
	if(!display) {
		if(displayString) {
			printf("error: could not open display %s\n", displayString);
		} else {
			printf("error: could not open display\n");
		}
		DoExit(1);
	}

	XSetErrorHandler(ErrorHandler);

	InitializeColors();
	InitializeFonts();
	InitializeCursors();

	rootWindow = RootWindow(display, DefaultScreen(display));
	rootWidth = DisplayWidth(display, DefaultScreen(display));
	rootHeight = DisplayHeight(display, DefaultScreen(display));

	attr.event_mask = SubstructureRedirectMask | SubstructureNotifyMask
		| PropertyChangeMask | ColormapChangeMask | ButtonPressMask
		| ButtonReleaseMask;
	XChangeWindowAttributes(display, rootWindow, CWEventMask, &attr);

	SetDefaultCursor(rootWindow);

	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);

#ifdef USE_SHAPE
	haveShape = XShapeQueryExtension(display, &shapeEvent, &temp);
	if(!haveShape) {
		Debug("No shape extension.");
	}
#endif

	InitializeLoadDisplay();
	CreateTray();
	InitializePager();

	/* Determine the alt key mask. */
	altCode = XKeysymToKeycode(display, XK_Alt_L);
	modmap = XGetModifierMapping(display);

	for(x = 0; x < 8 * modmap->max_keypermod; x++) {
		if(modmap->modifiermap[x] == altCode) {
			altMask = 1 << (x / modmap->max_keypermod);
		}
	}
	XFreeModifiermap(modmap);

	InitializeProtocols();

	LoadClients();

	XFlush(display);

	initializing = 0;

}

/****************************************************************************
 ****************************************************************************/
void HandleExit() {
	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);
	shouldExit = 1;
}

/****************************************************************************
 ****************************************************************************/
void Shutdown() {
	DestroyPager();
	DestroyRootMenu();
	DestroyLoadDisplay();
	DestroyTray();
	ReleaseClients();
	ReleaseCursors();
	ReleaseFonts();
	ReleaseColors();
	XCloseDisplay(display);
}


