/***************************************************************************
 * Menu functions display and handling functions.
 * Copyright (C) 2004 Joe Wingbermuehle
 ***************************************************************************/

#include "jwm.h"

static int ShowSubmenu(MenuType *menu, MenuType *parent,
	RunMenuCommandType runner, int x, int y);

static void CreateMenu(MenuType *menu, int x, int y);
static void HideMenu(MenuType *menu);
static void DrawMenu(MenuType *menu);

static int MenuLoop(MenuType *menu, RunMenuCommandType runner);
static int UpdateMotion(MenuType *menu, RunMenuCommandType runner,
	XMotionEvent *event);

static void UpdateMenu(MenuType *menu);
static MenuItemType *GetMenuItem(MenuType *menu, int index);

static int menuItemHeight;

/***************************************************************************
 ***************************************************************************/
void ShowMenu(MenuType *menu, RunMenuCommandType runner, int x, int y) {

	if(!GrabMouseForMenu()) {
		return;
	}

	menuItemHeight = menuFont->ascent + menuFont->descent + 1;

	ShowSubmenu(menu, NULL, runner, x, y);

	RefocusClient();
}

/***************************************************************************
 ***************************************************************************/
void DestroyMenu(MenuType *menu) {
	MenuItemType *np;
	if(menu) {
		while(menu->items) {
			np = menu->items->next;
			if(menu->items->name) {
				Release(menu->items->name);
			}
			if(menu->items->command) {
				Release(menu->items->command);
			}
			if(menu->items->submenu) {
				DestroyMenu(menu->items->submenu);
			}
			Release(menu->items);
			menu->items = np;
		}
		Release(menu);
	}
}

/***************************************************************************
 ***************************************************************************/
int ShowSubmenu(MenuType *menu, MenuType *parent, RunMenuCommandType runner,
	int x, int y) {

	int status;

	CreateMenu(menu, x, y);
	menu->parent = parent;

	status = MenuLoop(menu, runner);

	HideMenu(menu);

	if(!parent) {
		XUngrabPointer(display, CurrentTime);
		RefocusClient();
	}

	return status;
}

/***************************************************************************
 * Returns 0 if no selection was made or 1 if a selection was made.
 ***************************************************************************/
int MenuLoop(MenuType *menu, RunMenuCommandType runner) {
	XEvent event;
	MenuItemType *ip;
	int count;

	for(;;) {

		WaitForEvent(&event);

		switch(event.type) {
		case MotionNotify:
			switch(UpdateMotion(menu, runner, &event.xmotion)) {
			case 0: /* no selection */
				break;
			case 1: /* mouse left the menu */
				XPutBackEvent(display, &event);
				return 0;
			case 2: /* selection made */
				return 1;
			}
			break;
		case ButtonPress:
			if(menu->currentIndex >= 0) {
				count = 0;
				for(ip = menu->items; ip; ip = ip->next) {
					if(count == menu->currentIndex && ip->command) {
						(runner)(ip->command);
					}
					++count;
				}
			}
			return 1;
			break;
		default:
			break;
		}

	}
}

/***************************************************************************
 ***************************************************************************/
void CreateMenu(MenuType *menu, int x, int y) {
	MenuItemType *np;
	int temp;

	menu->lastIndex = -1;
	menu->currentIndex = -1;
	menu->width = 4;
	menu->height = 0;
	for(np = menu->items; np; np = np->next) {
		menu->height += menuItemHeight;
		if(np->name) {
			temp = XTextWidth(menuFont, np->name, strlen(np->name));
			if(temp > menu->width) {
				menu->width = temp;
			}
		}
	}
	menu->width += 8;

	if(x + menu->width > rootWidth) {
		x = rootWidth - menu->width;
	}
	temp = y;
	if(y + menu->height > rootHeight - TRAY_HEIGHT - 2) {
		y = rootHeight - TRAY_HEIGHT - menu->height - 2;
	}

	menu->x = x;
	menu->y = y;
	menu->parentOffset = temp - y;

	menu->window = XCreateSimpleWindow(display, rootWindow, x, y,
		menu->width, menu->height, 1, colors[COLOR_BORDER_OUTLINE],
		colors[COLOR_BORDER]);
	XMapRaised(display, menu->window);

	XFlush(display);

	XSetInputFocus(display, menu->window, RevertToPointerRoot, CurrentTime);
	menu->gc = XCreateGC(display, menu->window, 0, NULL);
	XSetFont(display, menu->gc, menuFont->fid);

	DrawMenu(menu);

}

/***************************************************************************
 ***************************************************************************/
void HideMenu(MenuType *menu) {
	XFreeGC(display, menu->gc);
	XDestroyWindow(display, menu->window);
}

/***************************************************************************
 ***************************************************************************/
void DrawMenu(MenuType *menu) {
	MenuItemType *np;
	int y;

	XSetForeground(display, menu->gc, colors[COLOR_BORDER]);
	XFillRectangle(display, menu->window, menu->gc,
		0, 0, menu->width, menu->height);

	XSetForeground(display, menu->gc, colors[COLOR_BORDER_TEXT]);
	XSetBackground(display, menu->gc, colors[COLOR_BORDER]);
	y = menuFont->ascent;
	for(np = menu->items; np; np = np->next) {
		if(np->name) {
			XDrawString(display, menu->window, menu->gc, 4, y,
				np->name, strlen(np->name));
		}
		y += menuItemHeight;
	}

}

/***************************************************************************
 * Return values:
 *  0 no selection.
 *  1 mouse left the menu.
 *  2 selection made on a submenu.
 ***************************************************************************/
int UpdateMotion(MenuType *menu, RunMenuCommandType runner,
	XMotionEvent *event) {

	MenuItemType *ip;
	MenuType *tp;
	int x, y;

	x = event->x - menu->x;
	y = event->y - menu->y;

	/* Update the selection on the current menu */
	if(x > 0 && y > 0 && x < menu->width && y < menu->height) {
		menu->currentIndex = y / menuItemHeight;
	} else {
		for(tp = menu->parent; tp; tp = tp->parent) {
			if(event->subwindow == tp->window) {
				if(y < menu->parentOffset
					|| y > menuItemHeight + menu->parentOffset) {
					return 1;
				}
			}
		}
		menu->currentIndex = -1;
	}
	if(menu->lastIndex != menu->currentIndex) {
		UpdateMenu(menu);
		menu->lastIndex = menu->currentIndex;
	}

	/* If the selected item is a submenu, show it. */
	ip = GetMenuItem(menu, menu->currentIndex);
	if(ip && ip->submenu) {
		if(ShowSubmenu(ip->submenu, menu, runner, menu->x + menu->width + 1,
			menu->y + menu->currentIndex * menuItemHeight)) {

			/* Item selected; destroy the menu tree. */
			return 2;

		} else {

			/* No selection made; redraw. */
			DrawMenu(menu);
			UpdateMenu(menu);

		}
	}

	return 0;

}

/***************************************************************************
 ***************************************************************************/
void UpdateMenu(MenuType *menu) {
	MenuItemType *ip;
	int yoffset;

	/* Clear the old selection. */
	ip = GetMenuItem(menu, menu->lastIndex);
	if(ip) {
		XSetForeground(display, menu->gc, colors[COLOR_BORDER]);
		XFillRectangle(display, menu->window, menu->gc,
			0, menu->lastIndex * menuItemHeight, menu->width,
			menuItemHeight + 2);
		XSetForeground(display, menu->gc, colors[COLOR_BORDER_TEXT]);
		XSetBackground(display, menu->gc, colors[COLOR_BORDER]);

		if(ip->name) {
			yoffset = menuFont->ascent + menu->lastIndex * menuItemHeight;
			XDrawString(display, menu->window, menu->gc, 4, yoffset,
				ip->name, strlen(ip->name));
		}
	}

	/* Highlight the new selection. */
	ip = GetMenuItem(menu, menu->currentIndex);
	if(ip) {
		XSetForeground(display, menu->gc, colors[COLOR_ACTIVE_BORDER]);
		XFillRectangle(display, menu->window, menu->gc,
			0, menu->currentIndex * menuItemHeight, menu->width,
			menuItemHeight);
		XSetForeground(display, menu->gc,
			colors[COLOR_ACTIVE_BORDER_TEXT]);
		XSetBackground(display, menu->gc, colors[COLOR_ACTIVE_BORDER]);

		if(ip->name) {
			yoffset = menuFont->ascent + menu->currentIndex * menuItemHeight;
			XDrawString(display, menu->window, menu->gc, 4, yoffset,
				ip->name, strlen(ip->name));
		}
	}

}

/***************************************************************************
 ***************************************************************************/
MenuItemType *GetMenuItem(MenuType *menu, int index) {
	MenuItemType *ip;

	if(index >= 0) {
		for(ip = menu->items; ip; ip = ip->next) {
			if(!index) {
				return ip;
			}
			--index;
		}
	} else {
		ip = NULL;
	}

	return ip;
}

