/****************************************************************************
 * Parser for the JWM XML configuration file.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static const char *restartCommand = "#restart";
static const char *restartName = "Restart";
static const char *exitCommand = "#exit";
static const char *exitName = "Exit";

static TokenNode *tokens;

static char *ReadFile(FILE *fd);

static void Parse(TokenNode *start);
static void ParseRootMenu(TokenNode *start);
static void ParseMenu(TokenNode *start, MenuType *menu);
static char *FindAttribute(AttributeNode *ap, const char *name);
static void ReleaseTokens(TokenNode *np);
static void SetDesktopCount(const char *value);

/****************************************************************************
 ****************************************************************************/
void ParseConfig(const char *fileName) {
	FILE *fd;
	char *buffer;

	fd = fopen(fileName, "r");
	if(!fd) {
		Warning("could not open configuration file: %s", fileName);
		return;
	}

	buffer = ReadFile(fd);
	fclose(fd);

	tokens = Tokenize(buffer);
	Release(buffer);
	Parse(tokens);
	ReleaseTokens(tokens);
}

/***************************************************************************
 ***************************************************************************/
void ReleaseTokens(TokenNode *np) {
	AttributeNode *ap;
	TokenNode *tp;

	while(np) {
		tp = np->next;

		while(np->attributes) {
			ap = np->attributes->next;
			if(np->attributes->name) {
				Release(np->attributes->name);
			}
			if(np->attributes->value) {
				Release(np->attributes->value);
			}
			Release(np->attributes);
			np->attributes = ap;
		}

		if(np->subnodeHead) {
			ReleaseTokens(np->subnodeHead);
		}

		if(np->value) {
			Release(np->value);
		}

		Release(np);
		np = tp;
	}

}

/***************************************************************************
 ***************************************************************************/
void Parse(TokenNode *start) {
	TokenNode *tp;
	int parsedRootMenu;

	if(!start) {
		return;
	}

	if(start->type == TOK_JWM) {
		parsedRootMenu = 0;
		for(tp = start->subnodeHead; tp; tp = tp->next) {
			switch(tp->type) {
			case TOK_BORDERFONT:
				SetBorderFont(tp->value);
				break;
			case TOK_TRAYFONT:
				SetTrayFont(tp->value);
				break;
			case TOK_MENUFONT:
				SetMenuFont(tp->value);
				break;
			case TOK_TEXTCOLOR:
				SetTextColor(tp->value);
				break;
			case TOK_ACTIVETEXTCOLOR:
				SetActiveTextColor(tp->value);
				break;
			case TOK_BORDERCOLOR:
				SetBorderColor(tp->value);
				break;
			case TOK_ACTIVEBORDERCOLOR:
				SetActiveBorderColor(tp->value);
				break;
			case TOK_DESKTOPCOUNT:
				SetDesktopCount(tp->value);
				break;
			case TOK_ROOTMENU:
				if(!parsedRootMenu) {
					ParseRootMenu(tp);
					parsedRootMenu = 1;
				} else {
					Warning("configuration error: "
						"root menu defined multiple times");
				}
				break;
			default:
				Warning("configuration error: invalid token: %d", tp->type);
				break;
			}
		}
	} else {
		Warning("configuration error: invalid start token: %d", start->type);
	}

}

/****************************************************************************
 ****************************************************************************/
void ParseRootMenu(TokenNode *start) {
	MenuType *menu;

	menu = Allocate(sizeof(MenuType));
	ParseMenu(start->subnodeHead, menu);
	SetRootMenu(menu);

}

/****************************************************************************
 ****************************************************************************/
void ParseMenu(TokenNode *start, MenuType *menu) {
	MenuItemType *current, *last;
	char *value;

	menu->items = NULL;
	current = NULL;
	last = NULL;
	while(start) {
		switch(start->type) {
		case TOK_MENU:

			current = Allocate(sizeof(MenuItemType));
			value = FindAttribute(start->attributes, "label");
			if(value) {
				current->name = Allocate(strlen(value) + 1);
				strcpy(current->name, value);
			} else {
				current->name = NULL;
			}
			current->command = NULL;
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			current->submenu = Allocate(sizeof(MenuType));
			ParseMenu(start->subnodeHead, current->submenu);

			break;
		case TOK_PROGRAM:

			current = Allocate(sizeof(MenuItemType));
			value = FindAttribute(start->attributes, "label");
			if(value) {
				current->name = Allocate(strlen(value) + 1);
				strcpy(current->name, value);
			} else if(start->value) {
				current->name = Allocate(strlen(start->value) + 1);
				strcpy(current->name, start->value);
			} else {
				current->name = NULL;
			}
			if(start->value) {
				current->command = Allocate(strlen(start->value) + 1);
				strcpy(current->command, start->value);
			} else {
				current->command = NULL;
			}
			current->submenu = NULL;
			current->next = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_SEPARATOR:

			current = Allocate(sizeof(MenuItemType));
			current->name = NULL;
			current->command = NULL;
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_EXIT:

			current = Allocate(sizeof(MenuItemType));
			current->name = Allocate(strlen(exitName) + 1);
			strcpy(current->name, exitName);
			current->command = Allocate(strlen(exitCommand) + 1);
			strcpy(current->command, exitCommand);
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_RESTART:

			current = Allocate(sizeof(MenuItemType));
			current->name = Allocate(strlen(restartName) + 1);
			strcpy(current->name, restartName);
			current->command = Allocate(strlen(restartCommand) + 1);
			strcpy(current->command, restartCommand);
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		default:
			Warning("configuration error: invalid node in menu: %d",
				start->type);
			break;
		}
		start = start->next;
	}
}

/***************************************************************************
 ***************************************************************************/
char *FindAttribute(AttributeNode *ap, const char *name) {

	while(ap) {
		if(!strcmp(name, ap->name)) {
			return ap->value;
		}
		ap = ap->next;
	}

	return NULL;
}

/***************************************************************************
 ***************************************************************************/
char *ReadFile(FILE *fd) {

	const int BLOCK_SIZE = 16;

	char *buffer;
	int len, max;
	int ch;

	len = 0;
	max = BLOCK_SIZE;
	buffer = Allocate(max + 1);

	for(;;) {
		ch = fgetc(fd);
		if(ch == EOF) {
			break;
		}
		buffer[len++] = ch;
		if(len >= max) {
			max += BLOCK_SIZE;
			buffer = Reallocate(buffer, max + 1);
		}
	}
	buffer[len] = 0;

	return buffer;
}

/****************************************************************************
 ****************************************************************************/
void SetDesktopCount(const char *value) {
	desktopCount = atoi(value);
	if(desktopCount <= 0) {
		desktopCount = 0;
	}
	if(desktopCount > MAX_DESKTOP_COUNT) {
		desktopCount = MAX_DESKTOP_COUNT;
	}
}


